/*
 * Original filename: mrc_readf.c (by Geert van Kempen)
 *
 * MEX-file to get the number of planes in a Biorad MRC-image.
 *
 * Usage:
 *   n = getmrcplanes('filename')
 * Planes start at 0!!!
 *
 * Cris Luengo
 * 21 June 1999
 * 9 July 1999: Fixed it to work properly.
 * 22 November 1999: Fixed the case-sensitivity of the extension check.
 *                   Added security of maximum number of characters copied when checking filename
 * 1 September 2000: Adapted this file from readmrcplane.
 */
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include "mex.h"

#define MRC_HEADER 76
#define MRC_NOTE 96
#define MRC_NOTE_TEXT 80
#define MRC_COOKIE 12345

#define MAX_FILE_NAME 256

#define MRC_CHARS_TO_SHORT(a,b) ((a)+256*(b))
#define MRC_CHARS_TO_INT(a,b,c,d) ((a)+256*(b)+65536*(c)+16777216*(d))

mxArray* getmrcplanes (const char *filename)
/* Note that filename has a maximum of 255 characters */
{
	FILE* fp;
	unsigned char header[MRC_HEADER], note[MRC_NOTE], note_text[MRC_NOTE_TEXT];
	char mrc_name[MAX_FILE_NAME+5], extension[MAX_FILE_NAME], buffer[MAX_FILE_NAME+50];
	char exts[2][5] = {".pic", ".PIC"};
	int num, note_type, note_next;
	bool found;
	int dims[2], numdims, planes;
	int sx, sy, sz, sp, cookie;
	void* ptr;
	mxArray* out;
	int ii, count;
	char* ext;

	/* Look for file, open it */
	found = false;
	if ((ext = strrchr (filename, '.')) != NULL) {
		strcpy (extension, ext);
		count = strlen (extension);
		for (ii = 0; ii < count; ii++) {
			extension[ii] = tolower(extension[ii]);
		}
		if ((strncmp (extension, ".pic", (size_t)4) == 0) && (fp = fopen (filename, "rb")))
			found = true;
	}
	if (!found) {
		num = 0;
		while ((num < 2) && !found ) {
			strcpy (mrc_name, filename);
			strcat (mrc_name, exts[num]);
			if (fp = fopen (mrc_name, "rb"))
				found = true;
			num++;
		}
	}
	if (!found) {
		sprintf (buffer, "Could not find Biorad MRC-image: \"%s\"", filename);
		mexErrMsgTxt (buffer);
		return (NULL);
	}

	/* read header */
	count = fread (header, sizeof (char), MRC_HEADER, fp);
	sx = MRC_CHARS_TO_SHORT (header[0], header[1]);
	sy = MRC_CHARS_TO_SHORT (header[2], header[3]);
	sz = MRC_CHARS_TO_SHORT (header[4], header[5]);
	sp = ((MRC_CHARS_TO_SHORT(header[14], header[15])) == 0) ? 2 : 1;
	cookie = MRC_CHARS_TO_SHORT(header[54], header[55]);
	if (sp != 1)
		mexErrMsgTxt ("The MRC file has 2 bytes per pixel. This is not yet supported.");

	/* check the cookie */
	if (cookie != MRC_COOKIE) {
		fclose (fp);
		sprintf (buffer, "The file \"%s\" has invalid MRC_COOKIE", filename);
		mexErrMsgTxt (buffer);
		return (NULL);
	}

	/* find number of planes */
	if (sz > 1) {
		planes = sz;
	}
	else {
		planes = 1;
	}

	fclose (fp);

	out = mxCreateDoubleMatrix(1,1,mxREAL);
	*mxGetPr(out) = (double)planes;

	return out;
}

/*
 * This is the MATLAB interface
 */
void mexFunction (int nlhs, mxArray* *plhs, int nrhs, const mxArray* *prhs) {

	char filename[MAX_FILE_NAME];

	/* Validate input */
	if (nrhs < 1)
		mexErrMsgTxt ("A file name is needed.");
	else if (nrhs > 2) {
		mexErrMsgTxt ("Too many input arguments.");
	}
	if (!mxIsChar (prhs[0]))
		mexErrMsgTxt ("A file name is needed.");
	mxGetString (prhs[0], filename, MAX_FILE_NAME);

	/* Validate output */
	if (nlhs > 1)
		mexErrMsgTxt ("Too many output arguments.");

	/* Do the stuff */
	plhs[0] = getmrcplanes (filename);
}
