/*
 * Original filename: mrc_readf.c (by Geert van Kempen)
 *
 * MEX-file to read the notes in a Biorad MRC-image.
 *
 * Usage:
 *   readmrcnotes('filename')
 *   data = readmrcnotes('filename')
 * The notes will be displayed if no output variable is given.
 * The output is a structure array with 3 elements: note type,
 * level and string.
 *
 * Cris Luengo
 * 21 June 1999
 * 9 July 1999: adapted this file from readmrc.
 * 12 July 1999: changed it to work properly.
 */
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include "mex.h"

#define MRC_HEADER 76
#define MRC_NOTE 96
#define MRC_NOTE_TEXT 80
#define MRC_COOKIE 12345

#define MAX_FILE_NAME 256

#define DEBUG_PRINT

#define MRC_CHARS_TO_SHORT(a,b) ((a)+256*(b))
#define MRC_CHARS_TO_INT(a,b,c,d) ((a)+256*(b)+65536*(c)+16777216*(d))

const char* element_names[] = {"type","level","string"};

mxArray* readmrc_notes (const char *filename, bool disp_notes)
/* Note that filename has a maximum of 255 characters */
{
	FILE* fp;
	unsigned char header[MRC_HEADER], note[MRC_NOTE], note_text[MRC_NOTE_TEXT];
	char mrc_name[MAX_FILE_NAME+5], extension[MAX_FILE_NAME], buffer[MAX_FILE_NAME+50];
	char exts[2][10] = {".pic", ".PIC"};
	int num, note_type, note_next, start_pos;
	bool found;
	int sx, sy, sz, sp, cookie;
	double* pr;
	mxArray* out = NULL;
	mxArray* field_value;
	int ii, count;
	char* ext;

	/* Look for file, open it */
	found = false;
	if ((ext = strrchr (filename, '.')) != NULL) {
		strcpy (extension, ext);
		count = strlen (extension);
		for (ii = 0; ii < count; ii++) {
			extension[ii] = tolower(extension[ii]);
		}
		if ((strncmp (extension, ".pic", (size_t)4) == 0) && (fp = fopen (filename, "rb")))
			found = true;
	}
	if (!found) {
		num = 0;
		while ((num < 2) && !found ) {
			strcpy (mrc_name, filename);
			strcat (mrc_name, exts[num]);
			if (fp = fopen (mrc_name, "rb"))
				found = true;
			num++;
		}
	}
	if (!found) {
		sprintf (buffer, "Could not find Biorad MRC-image: \"%s\"", filename);
		mexErrMsgTxt (buffer);
		return (NULL);
	}

	/* start processing */

	/* read header */
	count = fread (header, sizeof (char), MRC_HEADER, fp);
	sx = MRC_CHARS_TO_SHORT (header[0], header[1]);
	sy = MRC_CHARS_TO_SHORT (header[2], header[3]);
	sz = MRC_CHARS_TO_SHORT (header[4], header[5]);
	sp = ((MRC_CHARS_TO_SHORT(header[14], header[15])) == 0) ? 2 : 1;
	cookie = MRC_CHARS_TO_SHORT(header[54], header[55]);

	/* check the cookie */
	if (cookie != MRC_COOKIE) {
		fclose (fp);
		sprintf (buffer, "The file \"%s\" has invalid MRC_COOKIE", filename);
		mexErrMsgTxt (buffer);
		return (NULL);
	}

	/* Skip image data */
	fseek (fp, sp*sx*sy*sz, SEEK_CUR);

	/* read the note records, if present */
	if (disp_notes) {	/* For output */
		/*note_next = (getc (fp) == EOF) ? 0 : 1;*/
		count = fread (note, sizeof(char), MRC_NOTE, fp);
		if (count == MRC_NOTE)
			note_next = 1;
		count = 0;
		while (note_next) {
			note_next = MRC_CHARS_TO_INT (note[2], note[3], note[4], note[5]);
			note_type = MRC_CHARS_TO_SHORT(note[10], note[11]);
			strncpy (note_text, &note[16], MRC_NOTE_TEXT);
			mexPrintf ("Note (%d): %s\n", note_type, note_text);
			fread (note, sizeof(char), MRC_NOTE, fp);
			count++;
		}
	}
	else {	/* To write in an array */
		start_pos = ftell (fp);
		count = fread (note, sizeof(char), MRC_NOTE, fp);
		if (count != MRC_NOTE)
			goto mex_ending;
		count = 0; note_next = 1;
		while (note_next) {
			note_next = MRC_CHARS_TO_INT (note[2], note[3], note[4], note[5]);
			fread (note, sizeof(char), MRC_NOTE, fp);
			count++;
		}
		out = mxCreateStructMatrix (count, 1, 3, element_names);
		fseek (fp, start_pos, SEEK_SET);
		fread (note, sizeof(char), MRC_NOTE, fp);
		for (ii=0; ii<count; ii++) {
			field_value = mxCreateDoubleMatrix (1, 1, mxREAL);
			pr = mxGetPr (field_value);
			pr[0] = (double)MRC_CHARS_TO_SHORT(note[10], note[11]);
			mxSetFieldByNumber (out, ii, 0, field_value);
			field_value = mxCreateDoubleMatrix (1, 1, mxREAL);
			pr = mxGetPr (field_value);
			pr[0] = (double)MRC_CHARS_TO_SHORT(note[0], note[1]);
			mxSetFieldByNumber (out, ii, 1, field_value);
			strncpy (note_text, &note[16], MRC_NOTE_TEXT);
			field_value = mxCreateString (note_text);
			mxSetFieldByNumber (out, ii, 2, field_value);
			fread (note, sizeof(char), MRC_NOTE, fp);
		}
	}

mex_ending:
	fclose (fp);
	return out;
}

/*
 * This is the MATLAB interface
 */
void mexFunction (int nlhs, mxArray* *plhs, int nrhs, const mxArray* *prhs) {

	bool show_notes = false;
	char filename[256];

	/* Validate input */
	if (nrhs < 1)
		mexErrMsgTxt ("A file name is needed.");
	else if (nrhs > 1)
		mexErrMsgTxt ("Too many parameters.");

	if (!mxIsChar (prhs[0]))
		mexErrMsgTxt ("A file name is needed.");
	mxGetString (prhs[0], filename, 256);

	/* Validate output */
	if (nlhs > 1)
		mexErrMsgTxt ("Too many output arguments.");
	else if (nlhs == 0)
		show_notes = true;

	/* Do the stuff */
	if (show_notes)
		readmrc_notes (filename, true);
	else
		plhs[0] = readmrc_notes (filename, false);
}

